import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public
from nomad.datamodel.metainfo import common

m_package = Package(
    name='tinker_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='tinker.nomadmetainfo.json'))


class x_tinker_mdin_input_output_files(MCategory):
    '''
    Parameters of mdin belonging to x_tinker_section_control_parameters.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_tinker_mdin_input_output_files'))


class x_tinker_mdin_control_parameters(MCategory):
    '''
    Parameters of mdin belonging to x_tinker_section_control_parameters.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_tinker_mdin_control_parameters'))


class x_tinker_mdin_method(MCategory):
    '''
    Parameters of mdin belonging to section method.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_tinker_mdin_method'))


class x_tinker_mdout_single_configuration_calculation(MCategory):
    '''
    Parameters of mdout belonging to section_single_configuration_calculation.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_tinker_mdout_single_configuration_calculation'))


class x_tinker_mdout_method(MCategory):
    '''
    Parameters of mdin belonging to section method.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_tinker_mdout_method'))


class x_tinker_mdout_run(MCategory):
    '''
    Parameters of mdin belonging to settings run.
    '''

    m_def = Category(
        categories=[public.settings_run],
        a_legacy=LegacyDefinition(name='x_tinker_mdout_run'))


class x_tinker_mdin_run(MCategory):
    '''
    Parameters of mdin belonging to settings run.
    '''

    m_def = Category(
        categories=[public.settings_run],
        a_legacy=LegacyDefinition(name='x_tinker_mdin_run'))


class x_tinker_section_input_output_files(MSection):
    '''
    Section to store input and output file names
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_tinker_section_input_output_files'))


class x_tinker_section_control_parameters(MSection):
    '''
    Section to store the input and output control parameters
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_tinker_section_control_parameters'))

    x_tinker_inout_file_structure = Quantity(
        type=str,
        shape=[],
        description='''
        tinker input topology file.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_file_structure'))

    x_tinker_inout_file_trajectory = Quantity(
        type=str,
        shape=[],
        description='''
        tinker output trajectory file.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_file_trajectory'))

    x_tinker_inout_file_traj_coord = Quantity(
        type=str,
        shape=[],
        description='''
        tinker output trajectory file.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_file_traj_coord'))

    x_tinker_inout_file_traj_vel = Quantity(
        type=str,
        shape=[],
        description='''
        tinker output file for velocities in the trajectory.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_file_traj_vel'))

    x_tinker_inout_file_traj_force = Quantity(
        type=str,
        shape=[],
        description='''
        tinker output file for forces in the trajectory.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_file_traj_force'))

    x_tinker_inout_file_output_coord = Quantity(
        type=str,
        shape=[],
        description='''
        tinker output coordinates file.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_file_output_coord'))

    x_tinker_inout_file_output_vel = Quantity(
        type=str,
        shape=[],
        description='''
        tinker output velocities file.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_file_output_vel'))

    x_tinker_inout_file_output_force = Quantity(
        type=str,
        shape=[],
        description='''
        tinker output forces file.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_file_output_force'))

    x_tinker_inout_file_input_coord = Quantity(
        type=str,
        shape=[],
        description='''
        tinker input coordinates file.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_file_input_coord'))

    x_tinker_inout_file_input_vel = Quantity(
        type=str,
        shape=[],
        description='''
        tinker input velocities file.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_file_input_vel'))

    x_tinker_inout_file_restart_coord = Quantity(
        type=str,
        shape=[],
        description='''
        tinker restart coordinates file.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_file_restart_coord'))

    x_tinker_inout_file_restart_vel = Quantity(
        type=str,
        shape=[],
        description='''
        tinker restart velocities file.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_file_restart_vel'))

    x_tinker_inout_file_output_log = Quantity(
        type=str,
        shape=[],
        description='''
        tinker MD output log file.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_file_output_log'))

    x_tinker_inout_control_number_of_steps = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_number_of_steps'))

    x_tinker_inout_control_polar_eps = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_polar_eps'))

    x_tinker_inout_control_initial_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_initial_temperature'))

    x_tinker_inout_control_dielectric = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_dielectric'))

    x_tinker_inout_control_minimization = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_minimization'))

    x_tinker_inout_control_integrator = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_integrator'))

    x_tinker_inout_control_parameters = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_parameters'))

    x_tinker_inout_control_verbose = Quantity(
        type=bool,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_verbose'))

    x_tinker_inout_control_a_axis = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_a_axis'))

    x_tinker_inout_control_b_axis = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_b_axis'))

    x_tinker_inout_control_c_axis = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_c_axis'))

    x_tinker_inout_control_alpha = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_alpha'))

    x_tinker_inout_control_beta = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_beta'))

    x_tinker_inout_control_gamma = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_gamma'))

    x_tinker_inout_control_tau_pressure = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_tau_pressure'))

    x_tinker_inout_control_tau_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_tau_temperature'))

    x_tinker_inout_control_debug = Quantity(
        type=bool,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_debug'))

    x_tinker_inout_control_group = Quantity(
        type=bool,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_group'))

    x_tinker_inout_control_group_inter = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_group_inter'))

    x_tinker_inout_control_vib_roots = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_vib_roots'))

    x_tinker_inout_control_spacegroup = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_spacegroup'))

    x_tinker_inout_control_digits = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_digits'))

    x_tinker_inout_control_printout = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_printout'))

    x_tinker_inout_control_enforce_chirality = Quantity(
        type=bool,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_enforce_chirality'))

    x_tinker_inout_control_neighbor_list = Quantity(
        type=bool,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_neighbor_list'))

    x_tinker_inout_control_vdw_cutoff = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_vdw_cutoff'))

    x_tinker_inout_control_vdw_correction = Quantity(
        type=bool,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_vdw_correction'))

    x_tinker_inout_control_ewald = Quantity(
        type=bool,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_ewald'))

    x_tinker_inout_control_ewald_cutoff = Quantity(
        type=bool,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_ewald_cutoff'))

    x_tinker_inout_control_archive = Quantity(
        type=bool,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_archive'))

    x_tinker_inout_control_barostat = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_barostat'))

    x_tinker_inout_control_aniso_pressure = Quantity(
        type=bool,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_aniso_pressure'))

    x_tinker_inout_control_lights = Quantity(
        type=bool,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_lights'))

    x_tinker_inout_control_randomseed = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_randomseed'))

    x_tinker_inout_control_saddlepoint = Quantity(
        type=bool,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_saddlepoint'))

    x_tinker_inout_control_vdwtype = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_vdwtype'))

    x_tinker_inout_control_title = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_title'))

    x_tinker_inout_control_step_t = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_step_t'))

    x_tinker_inout_control_step_dt = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_step_dt'))

    x_tinker_inout_control_random_number_generator_seed = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_random_number_generator_seed'))

    x_tinker_inout_control_radiusrule = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_radiusrule'))

    x_tinker_inout_control_radiustype = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_radiustype'))

    x_tinker_inout_control_radiussize = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_radiussize'))

    x_tinker_inout_control_epsilonrule = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_epsilonrule'))

    x_tinker_inout_control_rattle = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_rattle'))

    x_tinker_inout_control_lambda = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_lambda'))

    x_tinker_inout_control_mutate = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_mutate'))

    x_tinker_inout_control_basin = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_basin'))

    x_tinker_inout_control_pme_grid = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_pme_grid'))

    x_tinker_inout_control_pme_order = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_pme_order'))

    x_tinker_inout_control_nstep = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_nstep'))

    x_tinker_inout_control_initial_configuration_file = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_initial_configuration_file'))

    x_tinker_inout_control_final_configuration_file = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_final_configuration_file'))

    x_tinker_inout_control_initial_trajectory_file = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_initial_trajectory_file'))

    x_tinker_inout_control_restart_file = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_restart_file'))

    x_tinker_inout_control_archive_file = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_archive_file'))

    x_tinker_inout_control_force_field_file = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_force_field_file'))

    x_tinker_inout_control_key_file = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_key_file'))

    x_tinker_inout_control_coordinate_file_list = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_coordinate_file_list'))

    x_tinker_inout_control_structure_file = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_structure_file'))

    x_tinker_inout_control_parameter_file = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_parameter_file'))

    x_tinker_inout_control_input_file = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_input_file'))

    x_tinker_inout_control_topology_file = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_topology_file'))

    x_tinker_inout_control_configuration_file = Quantity(
        type=str,
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_configuration_file'))

    x_tinker_inout_control_number_of_parameter_files = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_number_of_parameter_files'))

    x_tinker_inout_control_parameter_files = Quantity(
        type=str,
        shape=['x_tinker_inout_control_number_of_parameter_files'],
        description='''
        tinker running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_inout_control_parameter_files'))

    x_tinker_section_input_output_files = SubSection(
        sub_section=SectionProxy('x_tinker_section_input_output_files'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_tinker_section_input_output_files'))


class x_tinker_section_atom_to_atom_type_ref(MSection):
    '''
    Section to store atom label to atom type definition list
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_tinker_section_atom_to_atom_type_ref'))

    x_tinker_atom_to_atom_type_ref = Quantity(
        type=np.dtype(np.int64),
        shape=['number_of_atoms_per_type'],
        description='''
        Reference to the atoms of each atom type.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_atom_to_atom_type_ref'))


class x_tinker_section_single_configuration_calculation(MSection):
    '''
    section for gathering values for MD steps
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_tinker_section_single_configuration_calculation'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_tinker_atom_positions_image_index = Quantity(
        type=np.dtype(np.int32),
        shape=['number_of_atoms', 3],
        unit='dimensionless',
        description='''
        PBC image flag index.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_atom_positions_image_index'))

    x_tinker_atom_positions_scaled = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='dimensionless',
        description='''
        Position of the atoms in a scaled format [0, 1].
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_atom_positions_scaled'))

    x_tinker_atom_positions_wrapped = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='meter',
        description='''
        Position of the atoms wrapped back to the periodic box.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_atom_positions_wrapped'))

    x_tinker_lattice_lengths = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        Lattice dimensions in a vector. Vector includes [a, b, c] lengths.
        ''',
        categories=[public.configuration_core],
        a_legacy=LegacyDefinition(name='x_tinker_lattice_lengths'))

    x_tinker_lattice_angles = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        Angles of lattice vectors. Vector includes [alpha, beta, gamma] in degrees.
        ''',
        categories=[public.configuration_core],
        a_legacy=LegacyDefinition(name='x_tinker_lattice_angles'))

    x_tinker_dummy = Quantity(
        type=str,
        shape=[],
        description='''
        dummy
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_dummy'))

    x_tinker_mdin_finline = Quantity(
        type=str,
        shape=[],
        description='''
        finline in mdin
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_mdin_finline'))

    x_tinker_traj_timestep_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_traj_timestep_store'))

    x_tinker_traj_number_of_atoms_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_traj_number_of_atoms_store'))

    x_tinker_traj_box_bound_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_traj_box_bound_store'))

    x_tinker_traj_box_bounds_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_traj_box_bounds_store'))

    x_tinker_traj_variables_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_traj_variables_store'))

    x_tinker_traj_atoms_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_traj_atoms_store'))


class section_sampling_method(public.section_sampling_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_sampling_method'))

    x_tinker_barostat_target_pressure = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        MD barostat target pressure.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_barostat, public.settings_sampling],
        a_legacy=LegacyDefinition(name='x_tinker_barostat_target_pressure'))

    x_tinker_barostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD barostat relaxation time.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_barostat, public.settings_sampling],
        a_legacy=LegacyDefinition(name='x_tinker_barostat_tau'))

    x_tinker_barostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD barostat type, valid values are defined in the barostat_type wiki page.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_barostat, public.settings_sampling],
        a_legacy=LegacyDefinition(name='x_tinker_barostat_type'))

    x_tinker_integrator_dt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD integration time step.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_integrator, public.settings_sampling],
        a_legacy=LegacyDefinition(name='x_tinker_integrator_dt'))

    x_tinker_integrator_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD integrator type, valid values are defined in the integrator_type wiki page.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_integrator, public.settings_sampling],
        a_legacy=LegacyDefinition(name='x_tinker_integrator_type'))

    x_tinker_periodicity_type = Quantity(
        type=str,
        shape=[],
        description='''
        Periodic boundary condition type in the sampling (non-PBC or PBC).
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_integrator, public.settings_sampling],
        a_legacy=LegacyDefinition(name='x_tinker_periodicity_type'))

    x_tinker_langevin_gamma = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        Langevin thermostat damping factor.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_sampling, public.settings_thermostat],
        a_legacy=LegacyDefinition(name='x_tinker_langevin_gamma'))

    x_tinker_number_of_steps_requested = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of requested MD integration time steps.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_integrator, public.settings_sampling],
        a_legacy=LegacyDefinition(name='x_tinker_number_of_steps_requested'))

    x_tinker_thermostat_level = Quantity(
        type=str,
        shape=[],
        description='''
        MD thermostat level (see wiki: single, multiple, regional).
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_sampling, public.settings_thermostat],
        a_legacy=LegacyDefinition(name='x_tinker_thermostat_level'))

    x_tinker_thermostat_target_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='kelvin',
        description='''
        MD thermostat target temperature.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_sampling, public.settings_thermostat],
        a_legacy=LegacyDefinition(name='x_tinker_thermostat_target_temperature'))

    x_tinker_thermostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD thermostat relaxation time.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_sampling, public.settings_thermostat],
        a_legacy=LegacyDefinition(name='x_tinker_thermostat_tau'))

    x_tinker_thermostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD thermostat type, valid values are defined in the thermostat_type wiki page.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_sampling, public.settings_thermostat],
        a_legacy=LegacyDefinition(name='x_tinker_thermostat_type'))


class section_atom_type(common.section_atom_type):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_atom_type'))

    x_tinker_atom_name = Quantity(
        type=str,
        shape=[],
        description='''
        Atom name of an atom in topology definition.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_atom_name'))

    x_tinker_atom_type = Quantity(
        type=str,
        shape=[],
        description='''
        Atom type of an atom in topology definition.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_atom_type'))

    x_tinker_atom_element = Quantity(
        type=str,
        shape=[],
        description='''
        Atom type of an atom in topology definition.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_atom_element'))

    x_tinker_atom_type_element = Quantity(
        type=str,
        shape=[],
        description='''
        Element symbol of an atom type.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_atom_type_element'))

    x_tinker_atom_type_radius = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        van der Waals radius of an atom type.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_atom_type_radius'))

    number_of_atoms_per_type = Quantity(
        type=int,
        shape=[],
        description='''
        Number of atoms involved in this type.
        ''',
        a_legacy=LegacyDefinition(name='number_of_atoms_per_type'))


class section_interaction(common.section_interaction):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_interaction'))

    x_tinker_interaction_atom_to_atom_type_ref = Quantity(
        type=common.section_atom_type,
        shape=['number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type of each interaction atoms.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_interaction_atom_to_atom_type_ref'))

    x_tinker_number_of_defined_pair_interactions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of defined pair interactions (L-J pairs).
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_number_of_defined_pair_interactions'))

    x_tinker_pair_interaction_atom_type_ref = Quantity(
        type=common.section_atom_type,
        shape=['x_tinker_number_of_defined_pair_interactions', 'number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type for pair interactions.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_pair_interaction_atom_type_ref'))

    x_tinker_pair_interaction_parameters = Quantity(
        type=np.dtype(np.float64),
        shape=['x_tinker_number_of_defined_pair_interactions', 2],
        description='''
        Pair interactions parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_pair_interaction_parameters'))


class section_molecule_interaction(common.section_molecule_interaction):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_molecule_interaction'))

    x_tinker_molecule_interaction_atom_to_atom_type_ref = Quantity(
        type=common.section_atom_type,
        shape=['number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type of each molecule interaction atoms.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_molecule_interaction_atom_to_atom_type_ref'))

    x_tinker_number_of_defined_molecule_pair_interactions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of defined pair interactions within a molecule (L-J pairs).
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_number_of_defined_molecule_pair_interactions'))

    x_tinker_pair_molecule_interaction_parameters = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_defined_molecule_pair_interactions', 2],
        description='''
        Molecule pair interactions parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_pair_molecule_interaction_parameters'))

    x_tinker_pair_molecule_interaction_to_atom_type_ref = Quantity(
        type=common.section_atom_type,
        shape=['x_tinker_number_of_defined_pair_interactions', 'number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type for pair interactions within a molecule.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_pair_molecule_interaction_to_atom_type_ref'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_tinker_program_version_date = Quantity(
        type=str,
        shape=[],
        description='''
        Program version date.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_program_version_date'))

    x_tinker_parallel_task_nr = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Program task no.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_parallel_task_nr'))

    x_tinker_build_osarch = Quantity(
        type=str,
        shape=[],
        description='''
        Program Build OS/ARCH
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_build_osarch'))

    x_tinker_output_created_by_user = Quantity(
        type=str,
        shape=[],
        description='''
        Output file creator
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_output_created_by_user'))

    x_tinker_most_severe_warning_level = Quantity(
        type=str,
        shape=[],
        description='''
        Highest tinker warning level in the run.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_most_severe_warning_level'))

    x_tinker_program_build_date = Quantity(
        type=str,
        shape=[],
        description='''
        Program Build date
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_program_build_date'))

    x_tinker_program_citation = Quantity(
        type=str,
        shape=[],
        description='''
        Program citations
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_program_citation'))

    x_tinker_program_copyright = Quantity(
        type=str,
        shape=[],
        description='''
        Program copyright
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_program_copyright'))

    x_tinker_number_of_tasks = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of tasks in parallel program (MPI).
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_number_of_tasks'))

    x_tinker_program_module_version = Quantity(
        type=str,
        shape=[],
        description='''
        tinker program module version.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_program_module_version'))

    x_tinker_program_license = Quantity(
        type=str,
        shape=[],
        description='''
        tinker program license.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_program_license'))

    x_tinker_xlo_xhi = Quantity(
        type=str,
        shape=[],
        description='''
        test
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_xlo_xhi'))

    x_tinker_data_file_store = Quantity(
        type=str,
        shape=[],
        description='''
        Filename of data file
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_data_file_store'))

    x_tinker_program_working_path = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_program_working_path'))

    x_tinker_program_execution_host = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_program_execution_host'))

    x_tinker_program_execution_path = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_program_execution_path'))

    x_tinker_program_module = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_program_module'))

    x_tinker_program_execution_date = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_program_execution_date'))

    x_tinker_program_execution_time = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_program_execution_time'))

    x_tinker_mdin_header = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_mdin_header'))

    x_tinker_mdin_wt = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_mdin_wt'))

    x_tinker_section_control_parameters = SubSection(
        sub_section=SectionProxy('x_tinker_section_control_parameters'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_tinker_section_control_parameters'))


class section_topology(common.section_topology):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_topology'))

    x_tinker_input_units_store = Quantity(
        type=str,
        shape=[],
        description='''
        It determines the units of all quantities specified in the input script and data
        file, as well as quantities output to the screen, log file, and dump files.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_input_units_store'))

    x_tinker_data_bond_types_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_data_bond_types_store'))

    x_tinker_data_bond_count_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_data_bond_count_store'))

    x_tinker_data_angle_count_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_data_angle_count_store'))

    x_tinker_data_atom_types_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_data_atom_types_store'))

    x_tinker_data_dihedral_count_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_data_dihedral_count_store'))

    x_tinker_data_angles_store = Quantity(
        type=str,
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_data_angles_store'))

    x_tinker_data_angle_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_data_angle_list_store'))

    x_tinker_data_bond_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_data_bond_list_store'))

    x_tinker_data_dihedral_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_data_dihedral_list_store'))

    x_tinker_data_dihedral_coeff_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_data_dihedral_coeff_list_store'))

    x_tinker_masses_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_masses_store'))

    x_tinker_data_topo_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_data_topo_list_store'))

    x_tinker_section_atom_to_atom_type_ref = SubSection(
        sub_section=SectionProxy('x_tinker_section_atom_to_atom_type_ref'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_tinker_section_atom_to_atom_type_ref'))


class section_frame_sequence(public.section_frame_sequence):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_frame_sequence'))

    x_tinker_number_of_volumes_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of volumes in this sequence of frames, see
        x_tinker_frame_sequence_volume.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_number_of_volumes_in_sequence'))

    x_tinker_number_of_density_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of densities in this sequence of frames, see
        x_tinker_frame_sequence_density.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_number_of_density_in_sequence'))

    x_tinker_number_of_bond_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of bond_energies in this sequence of frames, see
        x_tinker_frame_sequence_bond_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_number_of_bond_energies_in_sequence'))

    x_tinker_number_of_virial_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of bond_energies in this sequence of frames, see
        x_tinker_frame_sequence_virial_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_number_of_virial_energies_in_sequence'))

    x_tinker_number_of_angle_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of angle_energies in this sequence of frames, see
        x_tinker_frame_sequence_angle_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_number_of_angle_energies_in_sequence'))

    x_tinker_number_of_proper_dihedral_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of proper_dihedral_energies in this sequence of frames, see
        x_tinker_frame_sequence_proper_dihedral_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_number_of_proper_dihedral_energies_in_sequence'))

    x_tinker_number_of_improper_dihedral_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of improper_dihedral_energies in this sequence of frames, see
        x_tinker_frame_sequence_improper_dihedral_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_number_of_improper_dihedral_energies_in_sequence'))

    x_tinker_number_of_cross_dihedral_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of cross_dihedral_energies in this sequence of frames, see
        x_tinker_frame_sequence_cross_dihedral_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_number_of_cross_dihedral_energies_in_sequence'))

    x_tinker_number_of_vdw_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of vdw_energies in this sequence of frames, see
        x_tinker_frame_sequence_vdw_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_number_of_vdw_energies_in_sequence'))

    x_tinker_number_of_boundary_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of boundary_energies in this sequence of frames, see
        x_tinker_frame_sequence_boundary_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_number_of_boundary_energies_in_sequence'))

    x_tinker_number_of_electrostatic_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of electrostatic_energies in this sequence of frames, see
        x_tinker_frame_sequence_electrostatic_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_number_of_electrostatic_energies_in_sequence'))

    x_tinker_number_of_total_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of total_energies in this sequence of frames, see
        x_tinker_frame_sequence_total_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_number_of_total_energies_in_sequence'))

    x_tinker_number_of_total_kinetic_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of total_kinetic_energies in this sequence of frames, see
        x_tinker_frame_sequence_total_kinetic_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_number_of_total_kinetic_energies_in_sequence'))

    x_tinker_number_of_misc_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of misc_energies in this sequence of frames, see
        x_tinker_frame_sequence_misc_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_number_of_misc_energies_in_sequence'))

    x_tinker_frame_sequence_density_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_tinker_number_of_density_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_tinker_frame_sequence_density values refers to. If not given it defaults to the
        trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_density_frames'))

    x_tinker_frame_sequence_density = Quantity(
        type=np.dtype(np.float64),
        shape=['x_tinker_number_of_density_in_sequence'],
        description='''
        Array containing the values of the density along this sequence of frames (i.e., a
        trajectory, a frame is one section_single_configuration_calculation). If not all
        frames have a value the indices of the frames that have a value are stored in
        frame_sequence_density_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_density'))

    x_tinker_frame_sequence_cross_dihedral_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_tinker_number_of_cross_dihedral_energy_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_tinker_frame_sequence_cross_dihedral_energy values refers to. If not given it
        defaults to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_cross_dihedral_energy_frames'))

    x_tinker_frame_sequence_cross_dihedral_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_tinker_number_of_cross_dihedral_energy_in_sequence'],
        description='''
        Array containing the values of the cross_dihedral_energy along this sequence of
        frames (i.e., a trajectory, a frame is one
        section_single_configuration_calculation). If not all frames have a value the
        indices of the frames that have a value are stored in
        frame_sequence_cross_dihedral_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_cross_dihedral_energy'))

    x_tinker_frame_sequence_improper_dihedral_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_tinker_number_of_improper_dihedral_energy_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_tinker_frame_sequence_improper_dihedral_energy values refers to. If not given it
        defaults to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_improper_dihedral_energy_frames'))

    x_tinker_frame_sequence_improper_dihedral_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_tinker_number_of_improper_dihedral_energy_in_sequence'],
        description='''
        Array containing the values of the improper_dihedral_energy along this sequence of
        frames (i.e., a trajectory, a frame is one
        section_single_configuration_calculation). If not all frames have a value the
        indices of the frames that have a value are stored in
        frame_sequence_improper_dihedral_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_improper_dihedral_energy'))

    x_tinker_frame_sequence_proper_dihedral_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_tinker_number_of_proper_dihedral_energy_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_tinker_frame_sequence_proper_dihedral_energy values refers to. If not given it
        defaults to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_proper_dihedral_energy_frames'))

    x_tinker_frame_sequence_proper_dihedral_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_tinker_number_of_proper_dihedral_energy_in_sequence'],
        description='''
        Array containing the values of the proper_dihedral_energy along this sequence of
        frames (i.e., a trajectory, a frame is one
        section_single_configuration_calculation). If not all frames have a value the
        indices of the frames that have a value are stored in
        frame_sequence_proper_dihedral_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_proper_dihedral_energy'))

    x_tinker_frame_sequence_bond_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_tinker_number_of_bond_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_tinker_frame_sequence_bond_energy values refers to. If not given it defaults to
        the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_bond_energy_frames'))

    x_tinker_frame_sequence_virial_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_tinker_number_of_virial_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_tinker_frame_sequence_virial_energy values refers to. If not given it defaults
        to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_virial_energy_frames'))

    x_tinker_frame_sequence_bond_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_tinker_number_of_bond_energies_in_sequence'],
        description='''
        Array containing the values of the bond_energy along this sequence of frames
        (i.e., a trajectory, a frame is one section_single_configuration_calculation). If
        not all frames have a value the indices of the frames that have a value are stored
        in frame_sequence_bond_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_bond_energy'))

    x_tinker_frame_sequence_virial_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_tinker_number_of_virial_energies_in_sequence'],
        description='''
        Array containing the values of the virial_energy along this sequence of frames
        (i.e., a trajectory, a frame is one section_single_configuration_calculation). If
        not all frames have a value the indices of the frames that have a value are stored
        in frame_sequence_virial_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_virial_energy'))

    x_tinker_frame_sequence_boundary_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_tinker_number_of_boundary_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_tinker_frame_sequence_boundary values refers to. If not given it defaults to the
        trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_boundary_frames'))

    x_tinker_frame_sequence_boundary = Quantity(
        type=np.dtype(np.float64),
        shape=['x_tinker_number_of_boundary_energies_in_sequence'],
        description='''
        Array containing the values of the boundary along this sequence of frames (i.e., a
        trajectory, a frame is one section_single_configuration_calculation). If not all
        frames have a value the indices of the frames that have a value are stored in
        frame_sequence_boundary_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_boundary'))

    x_tinker_frame_sequence_angle_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_tinker_number_of_angle_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_tinker_frame_sequence_angle_energy values refers to. If not given it defaults to
        the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_angle_energy_frames'))

    x_tinker_frame_sequence_angle_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_tinker_number_of_angle_energies_in_sequence'],
        description='''
        Array containing the values of the angle_energy along this sequence of frames
        (i.e., a trajectory, a frame is one section_single_configuration_calculation). If
        not all frames have a value the indices of the frames that have a value are stored
        in frame_sequence_angle_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_angle_energy'))

    x_tinker_frame_sequence_vdw_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_tinker_number_of_vdw_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_tinker_frame_sequence_vdw_energy values refers to. If not given it defaults to
        the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_vdw_energy_frames'))

    x_tinker_frame_sequence_vdw_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_tinker_number_of_vdw_energies_in_sequence'],
        description='''
        Array containing the values of the vdw_energy along this sequence of frames (i.e.,
        a trajectory, a frame is one section_single_configuration_calculation). If not all
        frames have a value the indices of the frames that have a value are stored in
        frame_sequence_vdw_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_vdw_energy'))

    x_tinker_frame_sequence_electrostatic_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_tinker_number_of_electrostatic_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_tinker_frame_sequence_electrostatic_energy values refers to. If not given it
        defaults to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_electrostatic_energy_frames'))

    x_tinker_frame_sequence_electrostatic_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_tinker_number_of_electrostatic_energies_in_sequence'],
        description='''
        Array containing the values of the electrostatic_energy along this sequence of
        frames (i.e., a trajectory, a frame is one
        section_single_configuration_calculation). If not all frames have a value the
        indices of the frames that have a value are stored in
        frame_sequence_electrostatic_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_electrostatic_energy'))

    x_tinker_frame_sequence_total_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_tinker_number_of_total_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_tinker_frame_sequence_total_energy values refers to. If not given it defaults to
        the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_total_energy_frames'))

    x_tinker_frame_sequence_total_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_tinker_number_of_total_energies_in_sequence'],
        description='''
        Array containing the values of the total_energy along this sequence of frames
        (i.e., a trajectory, a frame is one section_single_configuration_calculation). If
        not all frames have a value the indices of the frames that have a value are stored
        in frame_sequence_total_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_total_energy'))

    x_tinker_frame_sequence_total_kinetic_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_tinker_number_of_total_kinetic_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_tinker_frame_sequence_total_kinetic_energy values refers to. If not given it
        defaults to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_total_kinetic_energy_frames'))

    x_tinker_frame_sequence_total_kinetic_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_tinker_number_of_total_kinetic_energies_in_sequence'],
        description='''
        Array containing the values of the total_kinetic_energy along this sequence of
        frames (i.e., a trajectory, a frame is one
        section_single_configuration_calculation). If not all frames have a value the
        indices of the frames that have a value are stored in
        frame_sequence_total_kinetic_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_total_kinetic_energy'))

    x_tinker_frame_sequence_misc_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_tinker_number_of_misc_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_tinker_frame_sequence_misc_energy values refers to. If not given it defaults to
        the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_misc_energy_frames'))

    x_tinker_frame_sequence_misc_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_tinker_number_of_misc_energies_in_sequence'],
        description='''
        Array containing the values of the misc_energy along this sequence of frames
        (i.e., a trajectory, a frame is one section_single_configuration_calculation). If
        not all frames have a value the indices of the frames that have a value are stored
        in frame_sequence_misc_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_misc_energy'))

    x_tinker_frame_sequence_volume_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_tinker_number_of_volumes_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_tinker_frame_sequence_volume values refers to. If not given it defaults to the
        trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_volume_frames'))

    x_tinker_frame_sequence_volume = Quantity(
        type=np.dtype(np.float64),
        shape=['x_tinker_number_of_volumes_in_sequence'],
        description='''
        Array containing the values of the volume along this sequence of frames (i.e., a
        trajectory, a frame is one section_single_configuration_calculation). If not all
        frames have a value the indices of the frames that have a value are stored in
        frame_sequence_volume_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_tinker_frame_sequence_volume'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    x_tinker_section_single_configuration_calculation = SubSection(
        sub_section=SectionProxy('x_tinker_section_single_configuration_calculation'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_tinker_section_single_configuration_calculation'))


m_package.__init_metainfo__()



# pylint: disable=import-outside-toplevel,too-many-locals
import re
import urllib
from datetime import datetime
from typing import Union, List, Dict

from fastapi import HTTPException, Request

from optimade import __api_version__
from optimade.models import (
    ResponseMeta,
    EntryResource,
    EntryResponseMany,
    EntryResponseOne,
    ToplevelLinks,
    ReferenceResource,
    StructureResource,
)

from optimade.server.config import CONFIG
from optimade.server.entry_collections import EntryCollection
from optimade.server.exceptions import BadRequest
from optimade.server.query_params import EntryListingQueryParams, SingleEntryQueryParams


ENTRY_INFO_SCHEMAS = {
    "structures": StructureResource.schema,
    "references": ReferenceResource.schema,
}

BASE_URL_PREFIXES = {
    "major": f"/v{__api_version__.split('.')[0]}",
    "minor": f"/v{'.'.join(__api_version__.split('.')[:2])}",
    "patch": f"/v{__api_version__}",
}


def meta_values(
    url: str,
    data_returned: int,
    data_available: int,
    more_data_available: bool,
    **kwargs,
) -> ResponseMeta:
    """Helper to initialize the meta values"""
    from optimade.models import ResponseMetaQuery

    parse_result = urllib.parse.urlparse(url)

    # To catch all (valid) variations of the version part of the URL, a regex is used
    if re.match(r"/v[0-9]+(\.[0-9]+){,2}/.*", parse_result.path) is not None:
        url_path = re.sub(r"/v[0-9]+(\.[0-9]+){,2}/", "/", parse_result.path)
    else:
        url_path = parse_result.path

    return ResponseMeta(
        query=ResponseMetaQuery(representation=f"{url_path}?{parse_result.query}"),
        api_version=f"v{__api_version__}",
        time_stamp=datetime.utcnow(),
        data_returned=data_returned,
        more_data_available=more_data_available,
        provider=CONFIG.provider,
        data_available=data_available,
        implementation=CONFIG.implementation,
        **kwargs,
    )


def handle_response_fields(
    results: Union[List[EntryResource], EntryResource], exclude_fields: set
) -> dict:
    """Handle query parameter ``response_fields``

    It is assumed that all fields are under ``attributes``.
    This is due to all other top-level fields are REQUIRED in the response.

    :param exclude_fields: Fields under ``attributes`` to be excluded from the response.
    """
    if not isinstance(results, list):
        results = [results]

    new_results = []
    while results:
        entry = results.pop(0)
        new_entry = entry.dict(exclude_unset=True)
        for field in exclude_fields:
            if field in new_entry["attributes"]:
                del new_entry["attributes"][field]
        new_results.append(new_entry)
    return new_results


def get_included_relationships(
    results: Union[EntryResource, List[EntryResource]],
    ENTRY_COLLECTIONS: Dict[str, EntryCollection],
    include_param: List[str],
) -> Dict[str, List[EntryResource]]:
    """Filters the included relationships and makes the appropriate compound request
    to include them in the response.

    Parameters:
        results: list of returned documents.
        ENTRY_COLLECTIONS: dictionary containing collections to query, with key
            based on endpoint type.
        include_param: list of queried related resources that should be included in
            `included`.

    Returns:
        Dictionary with the same keys as ENTRY_COLLECTIONS, each containing the list
            of resource objects for that entry type.

    """
    from collections import defaultdict

    if not isinstance(results, list):
        results = [results]

    for entry_type in include_param:
        if entry_type not in ENTRY_COLLECTIONS and entry_type != "":
            raise BadRequest(
                detail=f"'{entry_type}' cannot be identified as a valid relationship type. "
                f"Known relationship types: {sorted(ENTRY_COLLECTIONS.keys())}"
            )

    endpoint_includes = defaultdict(dict)
    for doc in results:
        # convert list of references into dict by ID to only included unique IDs
        if doc is None:
            continue

        relationships = doc.relationships
        if relationships is None:
            continue

        relationships = relationships.dict()
        for entry_type in ENTRY_COLLECTIONS:
            # Skip entry type if it is not in `include_param`
            if entry_type not in include_param:
                continue

            entry_relationship = relationships.get(entry_type, {})
            if entry_relationship is not None:
                refs = entry_relationship.get("data", [])
                for ref in refs:
                    # could check here and raise a warning if any IDs clash
                    endpoint_includes[entry_type][ref["id"]] = ref

    included = {}
    for entry_type in endpoint_includes:
        compound_filter = " OR ".join(
            ['id="{}"'.format(ref_id) for ref_id in endpoint_includes[entry_type]]
        )
        params = EntryListingQueryParams(
            filter=compound_filter,
            response_format="json",
            response_fields=None,
            sort=None,
            page_limit=0,
            page_offset=0,
        )

        # still need to handle pagination
        ref_results, _, _, _ = ENTRY_COLLECTIONS[entry_type].find(params)
        included[entry_type] = ref_results

    # flatten dict by endpoint to list
    return [obj for endp in included.values() for obj in endp]


def get_base_url(parsed_url_request: urllib.parse.ParseResult) -> str:
    """Get base URL for current server

    Take the base URL from the config file, if it exists, otherwise use the request.
    """
    return (
        CONFIG.base_url
        if CONFIG.base_url
        else f"{parsed_url_request.scheme}://{parsed_url_request.netloc}"
    )


def get_entries(
    collection: EntryCollection,
    response: EntryResponseMany,
    request: Request,
    params: EntryListingQueryParams,
) -> EntryResponseMany:
    """Generalized /{entry} endpoint getter"""
    from optimade.server.routers import ENTRY_COLLECTIONS

    results, data_returned, more_data_available, fields = collection.find(params)

    include = []
    if getattr(params, "include", False):
        include.extend(params.include.split(","))
    included = get_included_relationships(results, ENTRY_COLLECTIONS, include)

    if more_data_available:
        # Deduce the `next` link from the current request
        parse_result = urllib.parse.urlparse(str(request.url))
        query = urllib.parse.parse_qs(parse_result.query)
        query["page_offset"] = int(query.get("page_offset", [0])[0]) + len(results)
        urlencoded = urllib.parse.urlencode(query, doseq=True)
        base_url = get_base_url(parse_result)

        links = ToplevelLinks(next=f"{base_url}{parse_result.path}?{urlencoded}")
    else:
        links = ToplevelLinks(next=None)

    if fields:
        results = handle_response_fields(results, fields)

    return response(
        links=links,
        data=results,
        meta=meta_values(
            url=str(request.url),
            data_returned=data_returned,
            data_available=len(collection),
            more_data_available=more_data_available,
        ),
        included=included,
    )


def get_single_entry(
    collection: EntryCollection,
    entry_id: str,
    response: EntryResponseOne,
    request: Request,
    params: SingleEntryQueryParams,
) -> EntryResponseOne:
    from optimade.server.routers import ENTRY_COLLECTIONS

    params.filter = f'id="{entry_id}"'
    results, data_returned, more_data_available, fields = collection.find(params)

    include = []
    if getattr(params, "include", False):
        include.extend(params.include.split(","))
    included = get_included_relationships(results, ENTRY_COLLECTIONS, include)

    if more_data_available:
        raise HTTPException(
            status_code=500,
            detail=f"more_data_available MUST be False for single entry response, however it is {more_data_available}",
        )

    links = ToplevelLinks(next=None)

    if fields and results is not None:
        results = handle_response_fields(results, fields)[0]

    return response(
        links=links,
        data=results,
        meta=meta_values(
            url=str(request.url),
            data_returned=data_returned,
            data_available=len(collection),
            more_data_available=more_data_available,
        ),
        included=included,
    )


def retrieve_queryable_properties(schema: dict, queryable_properties: list) -> dict:
    properties = {}
    for name, value in schema["properties"].items():
        if name in queryable_properties:
            if "$ref" in value:
                path = value["$ref"].split("/")[1:]
                sub_schema = schema.copy()
                while path:
                    next_key = path.pop(0)
                    sub_schema = sub_schema[next_key]
                sub_queryable_properties = sub_schema["properties"].keys()
                properties.update(
                    retrieve_queryable_properties(sub_schema, sub_queryable_properties)
                )
            else:
                properties[name] = {"description": value.get("description", "")}
                if "unit" in value:
                    properties[name]["unit"] = value["unit"]
                # All properties are sortable with the MongoDB backend.
                # While the result for sorting lists may not be as expected, they are still sorted.
                properties[name]["sortable"] = True
    return properties


def mongo_id_for_database(database_id: str, database_type: str) -> str:
    """Produce a MondoDB ObjectId for a database"""
    from bson.objectid import ObjectId

    oid = f"{database_id}{database_type}"
    if len(oid) > 12:
        oid = oid[:12]
    elif len(oid) < 12:
        oid = f"{oid}{'0' * (12 - len(oid))}"

    return str(ObjectId(oid.encode("UTF-8")))


def get_providers():
    """Retrieve Materials-Consortia providers (from https://providers.optimade.org/providers.json)"""
    import requests

    mat_consortia_providers = requests.get(
        "https://providers.optimade.org/providers.json"
    ).json()

    providers_list = []
    for provider in mat_consortia_providers.get("data", []):
        # Remove/skip "exmpl"
        if provider["id"] == "exmpl":
            continue

        provider.update(provider.pop("attributes"))

        # Add MongoDB ObjectId
        provider["_id"] = {
            "$oid": mongo_id_for_database(provider["id"], provider["type"])
        }

        providers_list.append(provider)

    return providers_list



#!/usr/bin/env python
# Authors: Franz Knuth & Fawzi Mohamed
from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import range
from builtins import object
import json
import numpy as np
import sys, re

def numpyDtypeForDtypeStr(dtypeStr):
    """returns the numpy dtype given the meta info dtype"""
    if dtypeStr == "f" or dtypeStr == "f64":
        return np.float64
    elif dtypeStr == "f32":
        return np.float32
    elif dtypeStr == "i" or dtypeStr == "i32":
        return np.int32
    elif dtypeStr == "i64":
        return np.int64
    elif dtypeStr == "b":
        return bool
    elif dtypeStr == "C":
        return object
    elif dtypeStr == "D":
        return object
    elif dtypeStr == "B":
        return object

def valueForStrValue(strValue, dtypeStr):
    """adds a value translating it from a string"""
    try:
        if dtypeStr[0] == "f":
            return float(strValue.replace("d","e").replace("D", "e"))
        elif dtypeStr[0] == "i":
            return int(strValue)
        elif dtypeStr[0] == "b":
            return (re.match("\s*(?:[nN][oO]?|\.?[fF](?:[aA][lL][sS][eE])?\.?|0)\s*$", strValue) is None)
        elif dtypeStr[0] == "B":
            return strValue # assumed to be base64 encoded
        elif dtypeStr[0] == "C":
            return strValue
        elif dtypeStr[0] == "D":
            return json.loads(strValue)
        else:
            raise Exception("unexpected dtypeStr %s" % (dtypeStr))
    except Exception as e:
        raise Exception("Error when converting %r to dtypeStr %r" % (strValue, dtypeStr), e)



"""
This module contains base classes for building parsers for the NoMaD project
with an object-oriented approach.
"""
from builtins import str
from builtins import object

import os
import copy
import numpy as np
import logging
from future.utils import with_metaclass
from abc import ABCMeta, abstractmethod

from nomadcore.unit_conversion import unit_conversion
from nomadcore.simple_parser import mainFunction
from nomadcore.local_backend import LocalBackend
from nomadcore.local_meta_info import load_metainfo
from nomadcore.caching_backend import CachingLevel
from nomadcore.simple_parser import extractOnCloseTriggers, extractOnOpenTriggers
from nomadcore.caching_backend import ActiveBackend

logger = logging.getLogger(__file__)


class ParserInterface(with_metaclass(ABCMeta, object)):
    """This class provides an interface for parsing. The end-user will
    typically only interact with an object that derives from this class. All
    the input is given to this class and the parsing is done by calling the
    parse() method. The parsing output is determined by the backend object that
    is given in the constructor as a dependency. If no backend is specified, a
    local backend that outputs results into a dictionary will be used.

    This class controls if a specific parser version should be instantiated and
    provides basic information about the parser, such as the parser-info
    -dictionary and the name of the metainfo file.

    Attributes:
        main_parser: Object that actually does the parsing and is
            setup by this class based on the given contents.
        parser_context: A wrapper class for all the parser related information.
            This is contructed here and then passed onto the different
            subparsers.
    """

    def __init__(
            self, metainfo_to_keep=None, backend=None, default_units=None,
            metainfo_units=None, debug=False, log_level=logging.ERROR, store=True):
        """
        Arguments:
            metainfo_to_keep: A list of metainfo names. This list is used to
                optimize the parsing process as optimally only the information
                relevant to these metainfos will be parsed.
            backend: An object to which the parser will give all the parsed data.
                The backend will then determine where and when to output that data.
        """
        try:
            logger.setLevel(log_level)
        except Exception:
            # might fail on custom loggers
            pass

        self.store = store
        self.debug = debug
        self.metainfo_env = None
        self.metaInfoEnv = None
        self.initialize(metainfo_to_keep, backend, default_units, metainfo_units)

    def setup_logger(self, new_logger):
        global logger
        logger = new_logger

        # tell tests about received logger
        new_logger.debug('received logger')

    def initialize(self, metainfo_to_keep, backend, default_units, metainfo_units):
        """Initialize the parser with the given environment.
        """
        self.parser_context = ParserContext()
        self.parser_context.metainfo_to_keep = metainfo_to_keep
        self.parser_context.file_service = FileService()
        self.parser_context.cache_service = CacheService()
        self.parser_context.parser_info = self.get_parser_info()
        self.main_parser = None
        self.backend = backend

        if self.metainfo_env is None:
            metainfo_filename = os.path.basename(self.get_metainfo_filename())
            from nomad.metainfo.legacy import python_package_mapping
            import importlib
            python_package_name, _ = python_package_mapping(metainfo_filename)
            python_package_name = '.'.join(python_package_name.split('.')[:-1])
            python_module = importlib.import_module(python_package_name)
            metainfo = getattr(python_module, 'm_env')
            self.metainfo_env = metainfo
            self.metaInfoEnv = self.metainfo_env.legacy_info_env()

        # Setup the metainfo environment.
        self.parser_context.metainfo_env = self.metaInfoEnv

        # Check the list of default units
        default_unit_map = {}
        if default_units is not None:
            for unit in default_units:
                dimension = unit_conversion.ureg(unit).dimensionality
                old_value = default_unit_map.get(str(dimension))
                if old_value is not None:
                    raise LookupError("You can only specify one default value per dimension in the 'default_units' list. There are two different units given for the dimension '{}'".format(dimension))
                default_unit_map[str(dimension)] = unit

        # Check the list of metainfo units
        if metainfo_units is not None:
            for metaname, unit in metainfo_units.items():

                # Check that the unit is OK
                unit_conversion.ureg(unit)

                # Check that the metaname is OK
                meta = self.metaInfoEnv.infoKinds.get(metaname)
                if meta is None:
                    raise KeyError("The metainfo name '{}' could not be found. Check for typos or try updating the metainfo repository.".format(metaname))

        # Save the default units
        self.parser_context.default_units = default_unit_map
        self.parser_context.metainfo_units = metainfo_units

    @abstractmethod
    def setup_version(self):
        """Deduce the version of the software that was used and setup a correct
        main parser. The main parser should subclass MainParser and be stored
        to the 'main_parser' attribute of this class. You can give the
        'parser_context' wrapper object in the main parser constructor to pass
        all the relevant data for it.
        """
        pass

    # @abstractmethod  # Keep this way for a file in order to avoid breaking
    def setup_main_parser(self, version_dictionary):
        """Setup the attribute 'main_parser' which contains the object that
        starts the parsing from the main file. The main parser should be
        decided based on the information contained on the given dictionary.

        Args:
            version_dictionary: Dictionary containing version information. It
            can include the version number, run type, etc.
        """
        pass

    def get_mainfile_regex(self):
        """Used to return the regular expression that is used to match the main
        file.

        Returns:
            str: regular expression as a string.
        """

    @abstractmethod
    def get_metainfo_filename(self):
        """This function should return the name of the metainfo file that is
        specific for this parser. When the parser is started, the metainfo
        environment is loaded from this file that is located within a separate
        repository (nomad-meta-info).

        Returns:
            A string containing the metainfo filename for this parser.
        """
        return None

    @abstractmethod
    def get_parser_info(self):
        """This function should return a dictionary containing the parser info.
        This info is printed to the JSON backend. it should be of the form:

            {'name': 'softwarename-parser', 'version': '1.0'}

        Returns:
            A dictionary containing information about this parser.
        """
        return None

    def parse(self, main_file):
        """Starts the actual parsing process, and outputs the results to the
        backend specified in the constructor.
        """
        # Initialize the backend.
        if self.backend is not None:
            self.parser_context.super_backend = self.backend(self.metainfo_env)
        else:
            from nomad.parsing.legacy import Backend
            self.parser_context.super_backend = Backend(self.metainfo_env)

        # Check that the main file exists
        if not os.path.isfile(main_file):
            raise ValueError(
                "Couldn't find the main file '{}'. Check that the path is valid "
                "and the file exists on this path.".format(main_file)
            )

        self.parser_context.main_file = main_file
        self.setup_version()
        if not self.main_parser:
            logger.error("The main parser has not been set up.")

        self.main_parser.parse(main_file)

        # If using a local backend, the results will have been saved to a
        # separate results dictionary which should be returned.
        try:
            return self.parser_context.super_backend.results
        except AttributeError:
            return None


class FileService(object):
    """Provides the interface to accessing files related to a calculation.

    Before you can use the service you must setup the root path, where every
    file related to this calculation resides. All queries filepaths will be
    handled as relative to this root folder.

    You can also setup ids that point to a certain path. This helps in querying
    the files as you don't have to remember the exact path and you can store
    paths for later use.

    Used to map file paths to certain ID's. This helps in setting up the
    Secondary parsers as you can associate file paths to simpler ID's that are
    easier to use.

    Attributes:
        root_folder: A path to the root folder
        _file_ids: A dictionary containing the mapping between file ids and filepaths
        _file_handles: A "private" dictionary containing the cached file handles
        _file_contents: A "private" dictionary containing the cached file contents
        _file_sizes: A "private" dictionary containing the cached file sizes
    """
    def __init__(self, root_folder=None):
        """
        Args:
            root_folder: a path to the root folder as a string.
        """
        self.map_id_to_path = {}
        if root_folder is not None:
            self.setup_root_folder(root_folder)

    def setup_root_folder(self, root_folder):
        """Setup the path to the root folder. Every filepath you set or get
        through this service should be relative to this root folder.
        """
        if os.path.isdir(root_folder):
            self.root_folder = root_folder
        else:
            raise IOError("Could not find the folder '{}'".format(root_folder))

    def get_absolute_path_to_file(self, relative_path):
        """
        Returns:
            If the given file is found, returns the absolute path to it. Return
            none if no file with the given path can be found.
        """
        path = os.path.join(self.root_folder, relative_path)
        if os.path.isfile(path):
            return path
        else:
            return None

    def get_file_by_id(self, file_id):
        """
        Returns:
            Handle to the file. Return none if no file with the given id has
            been set.
        """
        path = self.map_id_to_path.get(file_id)
        if path is None:
            logger.error("The id '{}' has no path associated with it.".format(file_id))
            return None
        else:
            return path

    def set_file_id(self, path, file_id):
        """Used to map a simple identifier string to a file path. When a file
        id has been setup, you can easily access the filepath anywhere in the
        code (FileService is shared via parser_context) by calling
        get_file_by_id().
        """
        if path is None:
            return None

        # Check if there is an old definition
        old = self.map_id_to_path.get(file_id)
        if old is not None:
            raise LookupError("The path '{}' is already associated with id '{}'".format(old, file_id))

        # Check that the file exists
        path = os.path.join(self.root_folder, path)
        if not os.path.isfile(path):
            logger.warning("Could not set the id for file '{}' as it cannot be found.".format(path))
            return None
        else:
            self.map_id_to_path[file_id] = path
            return path


class RegexService(object):
    """
    Stores basic regex definitions that can be reused on multiple parsers.
    """
    def __init__(self):
        self.float = "-?\d+\.\d+(?:[ED](?:\+|-)\d+)?"  # Regex for a floating point value
        self.int = "-?\d+"  # Regex for an integer
        self.word = "[\S]+"  # Regex for a single word. Can contain anything else but whitespace
        self.letter = "[^\W\d_]"  # Regex for a single alphabetical letter
        self.eol = "[^\n]+"  # Regex for a single alphabetical letter


class AbstractBaseParser(with_metaclass(ABCMeta, object)):
    """A base class for all objects that parse contents from files.

    When initialized with the parser_context, you can find the caching backend
    and the super backend as parameters of this object. These can then be used
    by the object to output parsing results.

    Attributes:
        file_path: Path to a file that is parsed by this class.
        parser_context: The ParserContext object that contains various in-depth
            information about the parsing environment.
        backend: The backend that will cache things according to the rules
            given in the main parser.
        super_backend: The final backend where things are forwarded to by the
            caching backend.
    """

    def __init__(self, parser_context):
        self.parser_context = parser_context
        self.backend = parser_context.caching_backend
        self.super_backend = parser_context.super_backend
        self.file_service = parser_context.file_service
        self.cache_service = parser_context.cache_service
        self.caching_levels = {}
        self.default_data_caching_level = CachingLevel.ForwardAndCache
        self.default_section_caching_level = CachingLevel.Forward
        self.on_close = {}
        self.on_open = {}

    def prepare(self):
        """This function will prepare everything for parsing.

        The onClose and onOpen callbacks are gathered and the ActiveBackend is
        initialized. You should call this function, or prepare things manually
        before trying to use push values to the backend.
        """
        # Gather the onClose and onOpen triggers
        for attr, callback in extractOnCloseTriggers(self).items():
            oldCallbacks = self.on_close.get(attr, None)
            if oldCallbacks:
                oldCallbacks.append(callback)
            else:
                self.on_close[attr] = [callback]
        for attr, callback in extractOnOpenTriggers(self).items():
            oldCallbacks = self.on_open.get(attr, None)
            if oldCallbacks:
                oldCallbacks.append(callback)
            else:
                self.on_open[attr] = [callback]

        # Initialize the Caching backend
        self.backend = ActiveBackend.activeBackend(
            metaInfoEnv=self.parser_context.metainfo_env,
            cachingLevelForMetaName=self.caching_levels,
            defaultDataCachingLevel=self.default_data_caching_level,
            defaultSectionCachingLevel=self.default_section_caching_level,
            onClose=self.on_close,
            onOpen=self.on_open,
            superBackend=self.parser_context.super_backend,
            default_units=self.parser_context.default_units,
            metainfo_units=self.parser_context.metainfo_units)

    def print_json_header(self):
        # self.super_backend.fileOut.write("[")
        # uri = "file://" + self.parser_context.main_file
        # self.backend.startedParsingSession(uri, self.parser_context.parser_info)
        pass

    def print_json_footer(self):
        # self.backend.finishedParsingSession("ParseSuccess", None)
        # self.super_backend.fileOut.write("]\n")
        pass

    @abstractmethod
    def parse(self, filepath):
        """Used to do the actual parsing. Inside this function you should push
        the parsing results into the Caching backend, or directly to the
        superBackend. You will also have to open new sections, but remember
        that certain sections may already be opened by other parsers.
        """


class MainHierarchicalParser(AbstractBaseParser):
    """A base class for all parsers that parse a file using a hierarchy of
    SimpleMatcher objects.

    Attributes:
        root_matcher: The SimpleMatcher object at the root of this parser.
            caching_levels: A dicionary containing the caching options
            that the ActiveBackend will use for individual metanames.

            Example:
                self.caching_levels = {
                    'section_XC_functionals': CachingLevel.ForwardAndCache,
                }
        caching_levels: Dictionary that stores the caching levels for
            individual metanames.
        default_data_caching_level: A default caching level for data, i.e.
            metainfo with kindStr=type_document_content or type_dimension
        default_section_caching_level: A default caching level for sections.
        onClose: a dictionary of onClose triggers. These are added to the
            triggers that are already present in the parser context object.
        cm: An optional CommonMatcher object that is used to store common
            onClose triggers, SimpleMatchers, caching levels, etc.

    """
    def __init__(self, parser_context):
        """
        Args:
            file_path: Path to the main file as a string.
            parser_context: The ParserContext object that contains various
                in-depth information about the parsing environment.
        """
        super(MainHierarchicalParser, self).__init__(parser_context)
        self.root_matcher = None
        self.regexs = RegexService()
        self.cm = None
        self.super_context = self

    def parse(self, filepath):
        """Starts the parsing. By default uses the SimpleParser scheme, if you
        want to use something else or customize the process just override this
        method in the subclass.
        """
        mainFunction(
                mainFileDescription=self.root_matcher,
                metaInfoEnv=self.parser_context.metainfo_env,
                parserInfo=self.parser_context.parser_info,
                # outF=self.parser_context.super_backend.fileOut,
                cachingLevelForMetaName=self.caching_levels,
                superContext=self.super_context,
                onClose=self.on_close,
                onOpen=self.on_open,
                default_units=self.parser_context.default_units,
                metainfo_units=self.parser_context.metainfo_units,
                superBackend=self.parser_context.super_backend,
                metaInfoToKeep=self.parser_context.metainfo_to_keep,
                mainFile=filepath)

    def startedParsing(self, fInName, parser):
        """Function is called when the parsing starts.

        Args:
            fInName: The name of the main file.
            parser: The SimpleParser object that has compiled the regex
                definitions inside the root_matcher attribute of this class.
        """
        self.parser_context.caching_backend = parser.backend
        self.parser_context.cache_service.parser_context = self.parser_context
        self.parser_context.super_backend = parser.backend.superBackend
        self.backend = self.parser_context.caching_backend
        self.super_backend = self.parser_context.super_backend
        if self.cm is not None:
            self.cm.backend = parser.backend

    def setup_common_matcher(self, common_matcher):
        """Used to setup the CommonMatcher object. This object will contain
        SimpleMatchers, onClose functions, caching levels, etc. that are common
        for many different HierarchicalParsers.

        Args:
            common_matcher: A CommonMatcher object from which to extract stuff.
        """
        self.cm = common_matcher
        self.on_close.update(common_matcher.getOnCloseTriggers())
        self.on_open.update(common_matcher.getOnOpenTriggers())
        self.caching_levels.update(common_matcher.caching_levels)


class CommonParser(object):
    """
    This class is used as a base class for objects that store and instantiate
    common parts of the hierarchical SimpleMatcher structure. The object can be
    shared for many MainHierarchicalParsers.
    """
    def __init__(self, parser_context):

        # Repeating regex definitions
        self.parser_context = parser_context
        self.backend = parser_context.caching_backend
        self.file_service = parser_context.file_service
        self.cache_service = parser_context.cache_service
        self.caching_levels = {}
        self.regexs = RegexService()

    def getOnCloseTriggers(self):
        """
        Returns:
            A dictionary containing a section name as a key, and a list of
            trigger functions associated with closing that section.
        """
        onClose = {}
        for attr, callback in extractOnCloseTriggers(self).items():
            onClose[attr] = [callback]
        return onClose

    def getOnOpenTriggers(self):
        """
        Returns:
            A dictionary containing a section name as a key, and a list of
            trigger functions associated with opening that section.
        """
        onOpen = {}
        for attr, callback in extractOnOpenTriggers(self).items():
            onOpen[attr] = [callback]
        return onOpen


class ParserContext(object):
    """A container class for storing and moving information about the parsing
    environment. A single ParserContext object is initialized by the Parser
    class, or it's subclass.
    """
    def __init__(self):
        self.main_file = None
        self.version_id = None
        self.metainfo_to_keep = None
        self.super_backend = None
        self.caching_backend = None
        self.default_units = None
        self.metainfo_units = None
        self.file_service = None
        self.metainfo_env = None
        self.parser_info = None
        self.cache_service = None


class CacheObject(object):
    """Wraps a value stored inside a CacheService.
    """
    def __init__(self, name, default_value=None, single=True, update=True):
        self.name = name
        self.update = update
        self.default_value = copy.copy(default_value)
        self._value = default_value
        self._single = single
        self._pushed = False
        self._updated = default_value is not None

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    def clear(self):
        self._updated = True
        self._value = self.default_value


class CacheService(object):
    """A class that can be used to store intermediate results in the parsing
    process. This is quite similar to the caching system that is used by the
    ActiveBackend object but this version also allows caching values that have
    no metadata associated with them. Also you can control how many times the
    data can be read and written from the cache.
    """
    def __init__(self, parser_context=None):
        self._cache = {}
        self.parser_context = parser_context

    def __getitem__(self, name):
        """Get the value identified by name. If the cachemode does not support
        getting the value, an exception is raised.

        returns:

        raises:
        """
        cache_object = self.get_cache_object(name)
        return cache_object.value

    def get_cache_object(self, name):

        cache_object = self._cache.get(name)
        if cache_object is None:
            logger.warning("The cache object for '{}' could not be found".format(name))
            return None
        return cache_object

    def __setitem__(self, name, value):
        """Used to set the value for an item. The CacheObject corresponding to
        the name has to be first created by using the function
        add_cache_object().
        """
        cache_object = self._cache[name]
        cache_object.value = value
        cache_object._updated = True

    def clear(self):
        """Frees all object from the cache.
        """
        for cache_object in self._cache.values():
            cache_object.clear()

    def add(self, name, value=None, single=True, update=True):
        """Used to add a cache object. Two cache objects with the same name are
        not allowed.

        Args:
            single (bool): If the value is allowed to be pushed only once.
            update (bool): If the value should be update before pushing.
        """
        if name in self._cache:
            raise LookupError("There already exists a cached value with the name '{}'. All keys in the CacheService should be unique.".format(name))
        cache_object = CacheObject(name, value, single, update)
        self._cache[name] = cache_object

    def check_push_allowed(self, cache_object):
        if cache_object._single:
            if cache_object._pushed:
                raise LookupError("The CacheService value '{}' has already been output to the backend. The CacheOutputMode does not allow this.".format(cache_object.name))
                return False

        if cache_object.update:
            if not cache_object._updated:
                raise LookupError("Could not push value '{}' to backend because it was not updated since last push.".format(cache_object.name))
                return False

        return True

    def addValue(self, metaname, name=None):
        """Pushes the value stored in the cache with the given name to
        the backend.

        If the name cannot be found in the dictionary, or the value associated
        with the name is None, nothing is pushed. This allows one to use the
        cache with a simple syntax without having to worry about the actual
        availability of the data.

        If the metaname property is not defined the value is pushed with the
        name that was used as key.
        """
        if name is None:
            name = metaname

        cache_object = self.get_cache_object(name)
        if cache_object is None or cache_object._value is None:
            logger.warning("The value for metaname '{}' was not set in the CacheService, and could not be pushed".format(name))
            return

        if self.check_push_allowed(cache_object):
            self.parser_context.caching_backend.addValue(metaname, cache_object.value)
            cache_object._pushed = True
            cache_object._updated = False

    def addRealValue(self, metaname, name=None, unit=None):
        """Pushes the real value stored in the cache with the given name to
        the backend. A unit conversion will be made if units are specified.

        If the name cannot be found in the dictionary, or the value associated
        with the name is None, nothing is pushed. This allows one to use the
        cache with a simple syntax without having to worry about the actual
        availability of the data.

        If the metaname property is not defined the value is pushed with the
        name that was used as key.
        """
        if name is None:
            name = metaname

        cache_object = self.get_cache_object(name)
        if cache_object is None or cache_object._value is None:
            logger.warning("The value for metaname '{}' was not set in the CacheService, and could not be pushed".format(name))
            return

        if self.check_push_allowed(cache_object):

            if cache_object.value is not None:
                self.parser_context.caching_backend.addRealValue(metaname, cache_object.value, unit=unit)
                cache_object._pushed = True
                cache_object._updated = False
            else:
                logger.warning("The value for metaname '{}' was not set in the CacheService, and could not be pushed".format(name))

    def addArrayValues(self, metaname, name=None, unit=None):
        """Pushes the array value stored in the cache with the given name to
        the backend. A unit conversion will be made if units are specified.

        If the name cannot be found in the dictionary, or the value associated
        with the name is None, nothing is pushed. This allows one to use the
        cache with a simple syntax without having to worry about the actual
        availability of the data.

        If the metaname property is not defined the value is pushed with the
        name that was used as key.
        """
        if name is None:
            name = metaname

        cache_object = self.get_cache_object(name)
        if cache_object is None or cache_object._value is None:
            logger.warning("The value for metaname '{}' was not set in the CacheService, and could not be pushed".format(name))
            return

        if self.check_push_allowed(cache_object):

            if cache_object.value is not None:
                self.parser_context.caching_backend.addArrayValues(metaname, np.array(cache_object.value), unit=unit)
                cache_object._pushed = True
                cache_object._updated = False
            else:
                logger.warning("The value for metaname '{}' was not set in the CacheService, and could not be pushed".format(name))



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Introduction
------------

This is about the *datamodel* that is used to represent NOMAD entries in our databases and
search engines. The respective data, also known as (repository) metadata is also part of
the NOMAD Archive and the datamodel is also defined based on the NOMAD Metainfo
(section *metadata*). It covers all information that users can search for and that can
be easily rendered on the GUI. The information is readily available through the
**repo** API.

.. image:: assets/data.png

See also the `datamodel section in the introduction <introduction.html#data-model>`_.

This module contains classes that allow to represent the core
nomad data entities (entries/calculations, users, datasets) on a high level of abstraction
independent from their representation in the different modules
:py:mod:`nomad.processing`, :py:mod:`nomad.parsing`, :py:mod:`nomad.search`, :py:mod:`nomad.app`.

Datamodel entities
------------------

The entities in the datamodel are defined as NOMAD Metainfo sections. They are treated
similarily to all Archive data. The entry/calculation datamodel data is created during
processing. It is not about representing every detail, but those parts that are directly involved in
api, processing, mirroring, or other 'infrastructure' operations.

The class :class:`User` is used to represent users and their attributes.

.. autoclass:: nomad.datamodel.User
    :members:

The class :class:`Dataset` is used to represent datasets and their attributes.

.. autoclass:: nomad.datamodel.Dataset
    :members:

The class :class:`MongoMetadata` is used to tag metadata stored in mongodb.

.. autoclass:: nomad.datamodel.MongoMetadata
    :members:

The class :class:`EntryMetadata` is used to represent all metadata about an entry.

.. autoclass:: nomad.datamodel.EntryMetadata
    :members:

Domains
------------------

The datamodel supports different *domains*. This means that most domain metadata of an
entry/calculation is stored in domain-specific sub sections of the :class:`EntryMetadata`
section. We currently have the following domain specific metadata classes/sections:

.. autoclass:: nomad.datamodel.dft.DFTMetadata
    :members:

.. autoclass:: nomad.datamodel.ems.EMSMetadata
    :members:

.. autoclass:: nomad.datamodel.OptimadeEntry
    :members:

.. automodule:: nomad.datamodel.encyclopedia
    :members:
'''
import sys

from nomad.metainfo import Environment

from .dft import DFTMetadata
from .ems import EMSMetadata
from .qcms import QCMSMetadata
from .datamodel import (
    Dataset, User, Author, EditableUserMetadata, UserProvidableMetadata, MongoMetadata,
    EntryMetadata, EntryArchive)
from .optimade import OptimadeEntry, Species
from .metainfo import m_env

m_env.m_add_sub_section(Environment.packages, sys.modules['nomad.datamodel.datamodel'].m_package)  # type: ignore
m_env.m_add_sub_section(Environment.packages, sys.modules['nomad.datamodel.dft'].m_package)  # type: ignore
m_env.m_add_sub_section(Environment.packages, sys.modules['nomad.datamodel.ems'].m_package)  # type: ignore
m_env.m_add_sub_section(Environment.packages, sys.modules['nomad.datamodel.qcms'].m_package)  # type: ignore
m_env.m_add_sub_section(Environment.packages, sys.modules['nomad.datamodel.encyclopedia'].m_package)  # type: ignore
m_env.m_add_sub_section(Environment.packages, sys.modules['nomad.datamodel.optimade'].m_package)  # type: ignore

domains = {
    'dft': {
        'metadata': DFTMetadata,
        'metainfo_all_package': 'common',
        'root_section': 'section_run'
    },
    'ems': {
        'metadata': EMSMetadata,
        'metainfo_all_package': 'common_experimental',
        'root_section': 'section_experiment'
    },
    'qcms': {
        'metadata': QCMSMetadata,
        'metainfo_all_package': 'general_qcms',
        'root_section': 'section_quantum_cms'
    }
}

root_sections = [domain['root_section'] for domain in domains.values()] + ['section_entry_info', 'OptimadeEntry', 'Workflow']



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flask_restplus import Resource
from flask import request

from nomad import config

from .api import api, url
from .common import base_request_args
from .models import json_api_single_response_model, base_endpoint_parser, json_api_single_response_model, Meta, json_api_list_response_model

ns = api.namespace('index/v1', description='This is the OPTiMaDe index for NOMAD\' implementations.')


@ns.route('/info')
class Info(Resource):
    @api.doc('index_info')
    @api.response(400, 'Invalid requests, e.g. bad parameter.')
    @api.expect(base_endpoint_parser, validate=True)
    @api.marshal_with(json_api_single_response_model, skip_none=True, code=200)
    def get(self):
        ''' Returns information relating to the API implementation- '''
        base_request_args()

        result = {
            'type': 'info',
            'id': '/',
            'attributes': {
                'api_version': '1.0.0',
                'available_api_versions': [{
                    'url': url(prefix='index'),
                    'version': '1.0.0'
                }],
                'formats': ['json'],
                'entry_types_by_format': {
                    'json': ['links', 'info']
                },
                'available_endpoints': ['links', 'info'],
                'is_index': True
            },
            'relationships': {
                'default': {
                    'data': {
                        'id': 'v1',
                        'type': 'links'
                    }
                }
            }
        }

        return dict(
            meta=Meta(query=request.url, returned=1),
            data=result
        ), 200


@ns.route('/links')
class Links(Resource):
    @api.doc('index_links')
    @api.response(400, 'Invalid requests, e.g. bad parameter.')
    @api.expect(base_endpoint_parser, validate=True)
    @api.marshal_with(json_api_list_response_model, skip_none=True, code=200)
    def get(self):
        ''' Returns information relating to the API implementation- '''
        base_request_args()

        result = [
            {
                "type": "child",
                "id": "v1",
                "attributes": {
                    "name": config.meta.name,
                    "description": config.meta.description,
                    "base_url": {
                        "href": url(version=None),
                    },
                    "homepage": config.meta.homepage
                }
            }
        ]

        return dict(
            meta=Meta(query=request.url, returned=1),
            data=result
        ), 200



import ase.io
import ase.io.formats
import mdtraj as md
import mdtraj.formats
import numpy as np
import logging
logger = logging.getLogger("nomad")


#===============================================================================
def iread(filename, file_format=None):
    """Generator function that is used to read an atomic configuration file (MD
    trajectory, geometry optimization, static snapshot) from a file one frame
    at a time. Only the xyz positions are returned from the file, and no unit
    conversion is done, so you have to be careful with units.

    By using a generator pattern we can avoid loading the entire trajectory
    file into memory. This function will instead load a chunk of the file into
    memory (with MDTraj you can decide the chunk size, with ASE it seems to
    always be one frame), and serve individual files from that chunk. Once the
    frames in one chunk are iterated, the chunk will be garbage collected and
    memory is freed.

    Args:
        filename: String for the file path.
        file_format: String for the file format. If not given the format is
            automatically detected from the extension.

    Yields:
        numpy array containing the atomic positions in one frame.

    """
    # If file format is not explicitly stated, determine the format from the
    # filename
    if file_format is None:
        file_format = filename.split(".")[-1]

    # Try to open the file with MDTraj first. With a brief inspection it seems
    # that MDTraj is better performance wise, because it can iteratively load a
    # "chunk" of frames, and still serve the individual frames one by one. ASE
    # on the other hand will iteratively read frames one by one (unnecessary
    # IO).
    mdtraj_chunk = 100  # How many frames MDTraj will load at once
    mdtraj_failed = False

    # Must use the low level MDTraj API to open files without topology.
    class_format_map = {
            "dcd": mdtraj.formats.DCDTrajectoryFile,
            "xyz": mdtraj.formats.XYZTrajectoryFile,
            "pdb": mdtraj.formats.PDBTrajectoryFile,
    }
    traj_class = class_format_map.get(file_format)
    if traj_class is not None:
        try:
            with traj_class(filename, mode="r") as f:
                empty = False
                while not empty:
                    data = f.read(mdtraj_chunk)
                    if isinstance(data, tuple):
                        positions = data[0]
                    else:
                        positions = data
                    if len(positions) == 0:
                        empty = True
                    else:
                        for pos in positions:
                            yield pos
        except IOError:
            logger.warning("MDTraj could not read the file '{}' with format '{}'. The contents might be malformed or wrong format used.".format(filename, file_format))
            return
    else:
        mdtraj_failed = True

    # If MDTraj didn't support the format, try ASE instead
    if mdtraj_failed:
        try:
            io = ase.io.formats.get_ioformat(file_format)
        except ValueError:
            logger.error("MDTraj could not read the file '{}' with format '{}'. If MDTraj is supposed to read this format, the contents might be malformed.".format(filename, file_format))
            return
        else:
            # Return the positions in a numpy array instead of an ASE Atoms object
            generator = ase.io.iread(filename, format=file_format)
            for atoms in generator:
                pos = atoms.positions
                yield pos



# Copyright 2016-2018 Mikkel Strange, Fawzi Mohamed
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from copy import copy
from scipy.io.netcdf import netcdf_file
from atkparser.parser_configurations import parse_configuration as p_conf
from atkparser.parser_calculator import parse_calculator as p_calc
import re

class X:
    def __init__(self, name=None):
        self.name = name

def has_data_with(what, f):
    for key in f.variables.keys():
        if what in key:
            return True
    return False

class Reader:
    def __init__(self, fname):
        self.f = netcdf_file(fname, 'r', mmap=True)
        self.CommonConcepts = X('CommonConcepts')
        self.CommonConcepts.Configurations = X('Configurations')

        gIDs = []
        for k in self.f.dimensions.keys():
            if '_gID' not in k:
                continue
            i = k.index('_gID')
            gID = int(k[i+4: i+7])
            if gID not in gIDs:
                gIDs.append(gID)

        self.gIDs = gIDs
        self.atk_version = self.f.version[4:].decode('utf-8')
        self.finger_prints = [x.split(':') for x in
                              self.f.fingerprint_table.\
                              decode('utf-8').split('#')][:-1]
        print(self.finger_prints)
        self.extract_common_concepts() #  atoms
        self.extract_calculator() # extract the calculator
        self.extract_total_energy() # look for total energy
        self.extract_results()  # look for results, forces, stress etc
        self.extract_wave_functions() # look for eigenvalues, wave functions
        self.extract_bandstructure()  # look for band structures

    def print_keys(self):
        """
            only for debuging
        """
        print('---dimensions---')
        for k in self.f.dimensions.keys():
            print(k)
        print('---variables---')
        for k in self.f.variables.keys():
            print(k)

    def extract_wave_functions(self):
        """ extract eigenvalues, occupations and wave_functions
        """
        self.wave_functions = X('wave_functions')

    def extract_calculator(self):
        self.calculators = []
        for configuration in self.configurations:
            self.calculators.append(p_calc(self.f, configuration))

    def extract_bandstructure(self):
        self.wave_functions = X('wave_functions')
        what = 'Bandstructure'
        if has_data_with(what, self.f) is False:
            return
        for key in self.f.variables.keys():
            if what in key:
                if 'route' in key:
                    self.wave_functions.route =\
                    self.f.variables[key][:].copy().tostring().decode('utf-8')
                    print(self.wave_functions.route)

    def extract_results(self):
        """
          try to read forces, stress and other stuff
        """
        self.results = X('results')


    def extract_total_energy(self):
        what = 'TotalEnergy'
        self.hamiltonian = X('hamiltonian')
        if has_data_with(what, self.f) is False:
            return
        for key in self.f.variables.keys():
            if what in key:
                if 'finger_print' in key:
                    self.hamiltonian.e_finger_print =\
                    self.f.variables[key][:].copy().tostring().decode('utf-8')
                elif 'Kinetic' in key:
                    self.hamiltonian.e_kin = self.f.variables[key][:].copy()[0]
                elif 'Exchange-Correlation' in key:
                    self.hamiltonian.e_xc = self.f.variables[key][:].copy()[0]
                elif 'External-Field' in key:
                    self.hamiltonian.e_external =\
                            self.f.variables[key][:].copy()[0]
                elif 'Electrostatic' in key:
                    self.hamiltonian.e_hartree = \
                            self.f.variables[key][:].copy()[0]
                elif 'Entropy-Term' in key:
                    self.hamiltonian.e_S = self.f.variables[key][:].copy()[0]
        ham = self.hamiltonian
        ham.e_tot = ham.e_kin + ham.e_xc + ham.e_hartree + ham.e_S

    def extract_common_concepts(self):
        #configurations could also be extraced from fingerprint_table
        self.configurations = []
        self.configurations_gIDs = []
        pm = re.compile('MoleculeConfiguration(?P<gid>_gID[0-9][0-9][0-9])$')
        pb = re.compile('BulkConfiguration(?P<gid>_gID[0-9][0-9][0-9])$')
        for key in self.f.variables.keys():
            for p in [pm, pb]:
                s = re.search(p, key)
                if s is not None:
                    self.configurations_gIDs.append(s.group('gid'))
                    self.configurations.append(s.group())
        self.configurations_atoms = []
        for conf in self.configurations:
            self.configurations_atoms.append(p_conf(self.f, conf))

if __name__ == '__main__':
    import sys
    fname = sys.argv[1]
    r = Reader(fname)
    #r.print_keys()
    #print(r.atoms)
    print(r.atk_version)
    print(r.hamiltonian.e_tot)
    print(r.calculators)
    print(r.configurations_atoms)
    print(r.configurations_atoms)




# Copyright 2016-2018 Mikkel Strange, Fawzi Mohamed
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from atkparser.periodic_table import things as ptab_ns
from atkparser.physical_quantities import things as physquan_ns
from atkparser.configurations import conf_types, things as confs_ns


def parse_configuration(fd, name, verbose=False):
    """ convert a nanolanguage python script into
    ASE list of atoms.

    Parameters:

        fd: netcdf_file handle
        name: str (i.e "BulkConfiguration_gID000")

    """
    code = fd.variables[name].data[:].copy()
    code = code.tostring().decode("utf-8")
    if 0:
        print('parsing code:\n------------')
        print(code)

    things = ptab_ns.copy()
    things.update(physquan_ns)
    things.update(confs_ns)
    exec(code, {}, things)
    for obj in things.values():
        for conf_type in conf_types:
            if isinstance(obj, confs_ns[conf_type]):
                return obj.atoms
    return -1


if __name__ == '__main__':
    from ase.visualize import view
    from scipy.io.netcdf import netcdf_file
    import sys
    fd = netcdf_file(sys.argv[1], 'r')
    name = 'BulkConfiguration_gID000'
    atoms = parse_configuration(fd, name)
    view(atoms)
#    configurations = {}
#    fp_gids = fd.fingerprint_table[:].decode('utf-8').split('#')
#    for c in fp_gids:
#        if len(c) > 0:
#            fingerprint, gID = c.split(':')[:2]
#            gID.strip()
#            for name in ['BulkConfiguration_' + gID,
#                         'MoleculeConfiguration_' + gID]:
#                if name in fd.variables.keys():
#                    configurations[name] = fingerprint
#
#    for name, fingerprint in configurations.items():
#        print(name+',', fingerprint)
#        atoms = parse_configuration(fd, name)
#        view(atoms)



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
This module contains functionality to use old 'legacy' NOMAD CoE parsers with the
new nomad@fairdi infrastructure. This covers aspects like the new metainfo, a unifying
wrapper for parsers, parser logging, and a parser backend.
'''

from typing import cast, Dict, List, Union, Any, Set, Iterable, Tuple, Type
import numpy as np
from pint.errors import UndefinedUnitError
import os.path
import importlib


from nomadcore.local_meta_info import loadJsonFile, InfoKindEl, InfoKindEnv

from nomad import utils
from nomad.units import ureg
from nomad.metainfo import (
    Definition, SubSection, Package, Quantity, Category, Section, Reference,
    Environment, MEnum, MSection, DefinitionAnnotation, MetainfoError, MSectionBound)

logger = utils.get_logger(__name__)


_ignored_packages = [
    'meta_types.nomadmetainfo.json',
    'repository.nomadmetainfo.json']


class LegacyDefinition(DefinitionAnnotation):

    def __init__(self, name: str):
        self.name = name


class LegacyPackage(LegacyDefinition):
    def __init__(self, name, python_module, python_path):
        super().__init__(name)

        self.python_module = python_module
        self.python_path = python_path


def def_name(definition):
    try:
        return definition.a_legacy.name
    except AttributeError:
        return definition.name


def normalize_name(name: str):
    return name.replace('.', '_').replace('-', '_')


def normalized_package_name(name: str):
    '''
    Transforms legacy metainfo '.nomadmetainfo.json' filenames into proper (python)
    identifier.
    '''
    name = name.replace('.nomadmetainfo.json', '')
    return normalize_name(name)


def python_package_mapping(metainfo_package_name: str) -> Tuple[str, str]:
    '''
    Compute the python package for the given metainfo package name. It returns
    a tuple containing a package name and a file path. The filepath denotes the file
    for this package within the nomad git project.
    '''
    prefix = metainfo_package_name.replace('.nomadmetainfo.json', '').split('.')[0]
    metainfo_package_name = normalized_package_name(metainfo_package_name)

    if prefix in ['common', 'general', 'public', 'dft', 'ems']:
        directory = 'nomad/datamodel/metainfo'
        python_package_name = 'nomad.datamodel.metainfo.%s' % metainfo_package_name

    else:
        parser_dir = prefix.replace('_', '-')
        prefix = prefix.replace('_', '')

        directory = 'dependencies/parsers/%s/%sparser/metainfo' % (parser_dir, prefix)
        python_package_name = '%sparser.metainfo.%s' % (prefix, metainfo_package_name)

    path = '%s/%s.py' % (directory, metainfo_package_name)

    return python_package_name, path


class LegacyMetainfoEnvironment(Environment):
    '''
    A metainfo environment with functions to create a legacy metainfo version of
    the environment.
    '''

    @staticmethod
    def from_legacy_package_path(path):
        metainfo_package_name = os.path.basename(path)
        package = metainfo_package_name
        if package.endswith('.nomadmetainfo.json'):
            package = package[:-19]
        if package.endswith('.json'):
            package = package[:-5]

        python_package_name, _ = python_package_mapping(package)
        python_package_name = '.'.join(python_package_name.split('.')[:-1])
        python_module = importlib.import_module(python_package_name)
        metainfo = getattr(python_module, 'm_env')

        return metainfo

    legacy_package_name = Quantity(type=str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__section_to_sub_section_name = None
        self.__legacy_names = None

    def from_legacy_name(self, name: str, section_cls: Type[MSectionBound]) -> MSectionBound:
        ''' Returns the definition of the given globally unique legacy metainfo name. '''
        if self.__legacy_names is None:
            self.__legacy_names = dict()
            for definition in self.m_all_contents():
                try:
                    if isinstance(definition, Section):
                        if definition.extends_base_section:
                            continue
                    legacy = definition.a_legacy
                    key = (legacy.name, definition.m_def.section_cls)
                    if key in self.__legacy_names:
                        raise MetainfoError('Legacy name %s is not globally unique' % legacy.name)
                    self.__legacy_names[key] = definition
                except AttributeError:
                    pass

        return self.__legacy_names.get((name, section_cls))

    @property
    def section_to_sub_section_name(self) -> Dict[str, str]:
        if self.__section_to_sub_section_name is not None:
            return self.__section_to_sub_section_name

        self.__section_to_sub_section_name = dict()
        for definition in self.m_all_contents():
            if definition.m_def == SubSection.m_def:
                self.__section_to_sub_section_name[definition.sub_section.name] = definition.name

        return self.__section_to_sub_section_name

    def legacy_info(self, definition: Definition, *args, **kwargs) -> InfoKindEl:
        ''' Creates a legacy metainfo object for the given definition. '''
        super_names: List[str] = list()
        result: Dict[str, Any] = dict(
            name=def_name(definition),
            description=definition.description,
            superNames=super_names)

        for category in definition.categories:
            super_names.append(def_name(category))

        if isinstance(definition, Section):
            sub_section_name = self.section_to_sub_section_name.get(definition.name, definition.name)
            result['kindStr'] = 'type_section'
            result['repeats'] = any(
                sub_section.repeats
                for sub_section in self.resolve_definitions(sub_section_name, SubSection))

            for sub_section in self.resolve_definitions(sub_section_name, SubSection):
                super_names.append(def_name(sub_section.m_parent_as(Definition)))

        elif isinstance(definition, Quantity):
            result['kindStr'] = 'type_document_content'
            result['shape'] = definition.shape
            dtype_str = None
            if definition.type == int:
                dtype_str = 'i'
            elif definition.type == float:
                dtype_str = 'f'
            elif definition.type == bool:
                dtype_str = 'b'
            elif definition.type == str:
                dtype_str = 'C'
            elif isinstance(definition.type, Reference):
                dtype_str = 'r'
                result['referencedSections'] = [
                    def_name(definition.type.target_section_def.m_resolved())]
            elif isinstance(definition.type, MEnum):
                dtype_str = 'C'
            elif type(definition.type) == np.dtype:
                dtype_str = definition.type.name[0]
            elif definition.type == Any:
                dtype_str = 'D'
            else:
                dtype_str = str(definition.type)
                # raise TypeError(
                #     'Unsupported quantity type %s in %s.' % (definition.type, definition))
            result['dtypeStr'] = dtype_str
            if definition.unit is not None:
                result['units'] = str(definition.unit)
            super_names.append(def_name(definition.m_parent_as(Definition)))

        elif isinstance(definition, Category):
            result['kindStr'] = 'type_abstract_document_content'

        package = cast(MSection, definition)
        while not isinstance(package, Package):
            package = package.m_parent

        result['package'] = package.name

        return InfoKindEl(*args, **result, **kwargs)

    def legacy_info_env(self, packages: List[Package] = None, *args, **kwargs) -> InfoKindEnv:
        ''' Creates a legacy metainfo environment with all definitions from the given packages. '''
        if packages is None:
            packages = self.packages

        env = InfoKindEnv(*args, **kwargs)
        for package in packages:
            for definition in package.all_definitions.values():
                if not (isinstance(definition, Section) and definition.extends_base_section):
                    env.addInfoKindEl(self.legacy_info(definition))

                if isinstance(definition, Section):
                    for quantity in definition.quantities:
                        env.addInfoKindEl(self.legacy_info(quantity))

        return env

    def to_legacy_dict(
            self, packages: List[Package] = None, description: str = None,
            *args, **kwargs) -> Dict[str, Any]:
        '''
        Creates a dictionary that can be serialized to a legacy metainfo definition file
        (*.nomadmetainfo.json).

        Arguments:
            package: Will add all definitions of these packages as actual definitions,
                all other packages will be added by import.
            description: The description for the legacy file. If None the description of
                the firs package will be used.
        '''
        if packages is None:
            packages = []

        definitions = []
        dependencies = []
        for package in self.packages:
            if package in packages:
                if description is None:
                    description = package.description

                for definition in package.all_definitions.values():
                    if not (isinstance(definition, Section) and definition.extends_base_section):
                        definitions.append(self.legacy_info(definition).toDict())

                    if isinstance(definition, Section):
                        for quantity in definition.quantities:
                            definitions.append(self.legacy_info(quantity).toDict())
            else:
                dependencies.append(package)

        return {
            'type': 'nomad_meta_info_1_0',
            'description': description,
            'dependencies': [
                {'relativePath': def_name(dependency)}
                for dependency in dependencies],
            'metaInfos': definitions
        }


class EnvironmentConversion:
    def __init__(self, legacy_env_or_path: Union[InfoKindEnv, str]):
        if isinstance(legacy_env_or_path, str):
            self.legacy_env, _ = loadJsonFile(filePath=legacy_env_or_path)

        else:
            self.legacy_env = cast(InfoKindEnv, legacy_env_or_path)

        self.__fix_legacy_super_names()

        self.package_conversions: Dict[str, PackageConversion] = {}

        for legacy_def in self.legacy_env.infoKindEls():
            if legacy_def.package in _ignored_packages:
                continue
            # legacy_def.package = normalized_package_name(legacy_def.package)
            package_conversion = self.package_conversions.get(legacy_def.package)
            if package_conversion is None:
                package_conversion = PackageConversion(self, legacy_def.package)
                self.package_conversions[legacy_def.package] = package_conversion

            package_conversion.legacy_defs.append(legacy_def)

        for package_conversion in self.package_conversions.values():
            package_conversion.create_definitions()

        for package_conversion in self.package_conversions.values():
            package_conversion.set_super_names()

        for package_conversion in self.package_conversions.values():
            package_conversion.init_definitions()

    def create_env(self) -> LegacyMetainfoEnvironment:
        env = LegacyMetainfoEnvironment()
        env.legacy_package_name = normalized_package_name(self.legacy_env.name)
        for package_conv in self.package_conversions.values():
            package = package_conv.package
            errors, warnings = package.m_all_validate()
            if len(errors) > 0:
                logger.error(
                    '%s. There are %d more errors in converted legacy package %s' %
                    (errors[0], len(errors) - 1, package))
            if len(warnings) > 0:
                logger.warn(
                    '%s. There are %d more warnings in converted legacy package %s' %
                    (warnings[0], len(warnings) - 1, package))
            env.m_add_sub_section(Environment.packages, package)
            package.init_metainfo()
        return env

    def __fix_legacy_super_names(self):

        def get_super_names(legacy_def: InfoKindEl, super_categories: Set[str] = None):
            super_section: str = None
            if super_categories is None:
                super_categories = set()

            for super_name in legacy_def.superNames:
                super_def = self.legacy_env.infoKindEl(super_name)

                if super_def.kindStr == 'type_section':
                    super_section = super_def.name

                elif super_def.kindStr == 'type_abstract_document_content':
                    super_categories.add(super_def.name)
                    super_super_section, _ = get_super_names(super_def, super_categories=super_categories)

                    if super_super_section is None:
                        pass

                    elif super_section is None:
                        super_section = super_super_section

                    elif super_section == super_super_section:
                        pass

                    else:
                        logger.error('conflicting parent sections %s, %s for %s' % (
                            super_section, super_def.name, legacy_def.name))

            return super_section, super_categories

        for legacy_def in self.legacy_env.infoKindEls():
            super_section, super_categories = get_super_names(legacy_def)

            if super_section is None:
                legacy_def.superNames = list(super_categories)

            else:
                legacy_def.superNames = [super_section] + list(super_categories)

    def resolve(self, name: str) -> Iterable[Definition]:
        for package_conversion in self.package_conversions.values():
            definition = package_conversion.package.all_definitions.get(name)
            if definition is not None:
                yield definition


class PackageConversion:

    def __init__(self, env_conversion: EnvironmentConversion, name: str):
        self.env_conversion = env_conversion
        self.legacy_defs: List[InfoKindEl] = []

        python_module, python_path = python_package_mapping(name)

        self.package = Package(
            name=normalize_name(name),
            a_legacy=LegacyPackage(name, python_module, python_path))

        self.quantities: Dict[str, Quantity] = {}

        self.logger = logger.bind(package=name)

    def create_definitions(self):
        for legacy_def in self.legacy_defs:
            name = normalize_name(legacy_def.name)

            if legacy_def.kindStr == 'type_abstract_document_content':
                self.package.m_create(
                    Category, name=name, a_legacy=LegacyDefinition(name=legacy_def.name))

            elif legacy_def.kindStr == 'type_section':
                self.package.m_create(
                    Section, name=name,
                    a_legacy=LegacyDefinition(name=legacy_def.name))

            elif legacy_def.kindStr in ['type_dimension', 'type_document_content']:
                definition = Quantity(
                    name=name,
                    a_legacy=LegacyDefinition(name=legacy_def.name))
                self.quantities[name] = (definition)

            else:
                logger.error('unknown kindStr %s for %s' % (legacy_def.kindStr, name))

    def __resolve(self, name: str, create_extends: bool = False):
        definition: Definition = self.package.all_definitions.get(name)
        if definition is None:
            definition = self.quantities.get(name)

        if definition is not None:
            if not (isinstance(definition, Section) and definition.extends_base_section) or create_extends:
                return definition

        for definition in self.env_conversion.resolve(name):
            if isinstance(definition, Section) and definition.extends_base_section:
                continue

            if create_extends and isinstance(definition, Section):
                extending_def = self.package.m_create(
                    Section, name=definition.name,
                    a_legacy=LegacyDefinition(name=definition.a_legacy.name))
                extending_def.base_sections = [definition]
                extending_def.extends_base_section = True
                return extending_def

            return definition

        assert False, 'definition %s must be created now' % name

    def set_super_names(self):
        for legacy_def in self.legacy_defs:
            name = normalize_name(legacy_def.name)
            definition = self.__resolve(name)
            assert definition is not None, 'definition %s must exist' % name

            if isinstance(definition, Section):
                parent_section: Section = None
                for super_name in legacy_def.superNames:
                    super_def = self.__resolve(normalize_name(super_name), create_extends=True)
                    if isinstance(super_def, Section):
                        parent_section = cast(Section, super_def)

                if parent_section is not None:
                    sub_section = parent_section.m_create(
                        SubSection, name=definition.name,
                        a_legacy=LegacyDefinition(name=legacy_def.name))
                    sub_section.sub_section = definition
                    sub_section.repeats = legacy_def.repeats is None or legacy_def.repeats

            if isinstance(definition, Quantity):
                parent_section: Section = None
                for super_name in legacy_def.superNames:
                    super_def = self.__resolve(normalize_name(super_name), create_extends=True)
                    if isinstance(super_def, Section):
                        parent_section = cast(Section, super_def)

                parent_section.m_add_sub_section(Section.quantities, definition)

    def init_definitions(self):
        for legacy_def in self.legacy_defs:
            name = normalize_name(legacy_def.name)
            definition = self.__resolve(name)
            assert definition is not None, 'definition %s must exist' % name
            logger = self.logger.bind(definition=definition.name)

            # common properties
            if legacy_def.description is not None and legacy_def.description.strip() != '':
                definition.description = legacy_def.description

            if isinstance(definition, Definition):
                # deal with categories
                categories: List[Category] = []
                for super_name in legacy_def.superNames:
                    super_def = self.__resolve(normalize_name(super_name))
                    if isinstance(super_def, Category):
                        categories.append(super_def)

                definition.categories = categories

            if isinstance(definition, Quantity):
                # type
                referenced_sections = legacy_def.extra_args.get('referencedSections', [])
                if len(referenced_sections) == 1:
                    referenced_section = self.__resolve(referenced_sections[0])
                    if referenced_section is None:
                        logger.error('could not find referencedSection %s of %s' % (
                            referenced_sections[0], definition.name))
                        definition.type = int
                    else:
                        definition.type = Reference(referenced_section)

                elif len(referenced_sections) > 1:
                    logger.error(
                        'higher dimensional references not yet supported: %s' % name)
                    definition.type = np.dtype(int)

                elif legacy_def.kindStr == 'type_dimension':
                    definition.type = int
                elif legacy_def.dtypeStr == 'D':
                    definition.type = Any
                elif legacy_def.dtypeStr == 'C':
                    definition.type = str
                elif legacy_def.dtypeStr == 'r':
                    logger.error('r typed quantity %s  doesn\'t have referencedSections' % name)
                    definition.type = int
                elif legacy_def.dtypeStr == 'b':
                    definition.type = bool
                elif legacy_def.dtypeStr == 'i64':
                    definition.type = np.dtype(np.int64)
                elif legacy_def.dtypeStr == 'f':
                    definition.type = np.dtype(np.float64)
                else:
                    definition.type = np.dtype(legacy_def.dtypeStr)

                # shapes
                legacy_shape = legacy_def.shape
                if legacy_shape is None:
                    legacy_shape = []

                definition.shape = legacy_shape
                if len(definition.shape) > 1 and definition.type == str:
                    # Usually only np types have higher shapes in old metainfo;
                    # str is one exception.
                    definition.type = np.dtype('U')

                # units
                if legacy_def.units is not None:
                    try:
                        definition.unit = ureg.parse_units(legacy_def.units)
                    except UndefinedUnitError:
                        logger.error('unknown unit %s' % legacy_def.units)
                    except ValueError as e:
                        logger.error('cannot parse unit %s' % legacy_def.units, exc_info=e)


def convert(metainfo_path: str) -> LegacyMetainfoEnvironment:
    return EnvironmentConversion(metainfo_path).create_env()


def generate_metainfo_code(metainfo_env: LegacyMetainfoEnvironment):
    '''
    Generates python code with metainfo definitions for all packages in the given
    environement

    Arguments:
        env: The metainfo environment.
        python_package_path: An optional directory path. The directory must exist. Default
            is the working directory. The path will be used to form the module prefix
            for generated Python modules.
    '''
    from jinja2 import Environment as JinjaEnvironment, PackageLoader, select_autoescape
    import textwrap

    def format_description(description, indent=0, width=90):
        paragraphs = [paragraph.strip() for paragraph in description.split('\n')]

        def format_paragraph(paragraph, first):
            lines = textwrap.wrap(text=paragraph, width=width - indent * 4)
            lines = [line.replace('\\', '\\\\') for line in lines]
            return textwrap.indent(
                '\n'.join(lines), ' ' * 4 * indent, lambda x: not (first and x.startswith(lines[0])))

        return '\n\n'.join([
            format_paragraph(p, i == 0)
            for i, p in enumerate(paragraphs) if p != ''])

    def format_type(pkg, mi_type):
        if type(mi_type) == np.dtype:
            if mi_type == np.dtype('U'):
                return 'np.dtype(\'U\')'

            return 'np.dtype(np.%s)' % mi_type

        if mi_type in [int, float, str, bool]:
            return mi_type.__name__

        if isinstance(mi_type, Reference):
            if pkg == mi_type.target_section_def.m_parent:
                return "Reference(SectionProxy('%s'))" % mi_type.target_section_def.name

            else:
                python_module = mi_type.target_section_def.m_parent.a_legacy.python_module
                return '%s.%s' % (python_module.split('.')[-1], mi_type.target_section_def.name)

        else:
            return str(mi_type)

    def format_unit(unit):
        return "'%s'" % unit

    def format_definition_refs(pkg, definitions):
        def format_definition_ref(definition: Definition):
            if pkg == definition.m_parent:
                return definition.name
            else:
                python_module = definition.m_parent.a_legacy.python_module
                return '%s.%s' % (python_module.split('.')[-1], definition.name)

        return ', '.join([format_definition_ref(definition) for definition in definitions])

    def fromat_package_import(pkg):
        python_module = pkg.a_legacy.python_module
        modules = python_module.split('.')
        return 'from %s import %s' % ('.'.join(modules[:-1]), modules[-1])

    def order_categories(categories):
        return sorted(categories, key=lambda c: len(c.categories))

    env = JinjaEnvironment(
        loader=PackageLoader('nomad.metainfo', 'templates'),
        autoescape=select_autoescape(['python']))
    env.globals.update(
        order_categories=order_categories,
        format_description=format_description,
        format_type=format_type,
        format_unit=format_unit,
        format_definition_refs=format_definition_refs,
        fromat_package_import=fromat_package_import)

    for package in metainfo_env.packages:
        path = package.a_legacy.python_path
        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))

        with open(path, 'wt') as f:
            code = env.get_template('package.j2').render(pkg=package)
            code = '\n'.join([
                line.rstrip() if line.strip() != '' else ''
                for line in code.split('\n')])
            f.write(code)

    _, path = python_package_mapping(metainfo_env.legacy_package_name)
    with open(os.path.join(os.path.dirname(path), '__init__.py'), 'wt') as f:

        code = env.get_template('environment.j2').render(env=metainfo_env)
        code = '\n'.join([
            line.rstrip() if line.strip() != '' else ''
            for line in code.split('\n')])
        f.write(code)



# Copyright 2016-2018 Lauri Himanen, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import absolute_import
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.baseclasses import MainHierarchicalParser
import numpy as np
from .commonparser import CPMDCommonParser
import re
import logging
LOGGER = logging.getLogger("nomad")


class CPMDVibrationParser(MainHierarchicalParser):
    """The main parser class that is called for all run types. Parses the CPMD
    output file.
    """
    def __init__(self, file_path, parser_context):
        """
        """
        super(CPMDVibrationParser, self).__init__(file_path, parser_context)
        self.setup_common_matcher(CPMDCommonParser(parser_context))
        self.n_frames = 0
        self.sampling_method_gid = None
        self.frame_refs = []
        self.energies = []

        #=======================================================================
        # Main structure
        self.root_matcher = SM("",
            forwardMatch=True,
            sections=['section_run', "section_method"],
            subMatchers=[
                self.cm.header(),
                self.cm.method(),
                self.cm.atoms(),
                self.cm.cell(),
                self.cm.initialization(),
                self.cm.footer(),
            ]
        )

    #=======================================================================
    # onClose triggers
    def onClose_section_system(self, backend, gIndex, section):
        self.cache_service.addArrayValues("atom_labels")
        self.cache_service.addArrayValues("simulation_cell", unit="bohr")
        self.cache_service.addValue("number_of_atoms")

    #=======================================================================
    # adHoc
